﻿//*********************************************************//
//    Copyright (c) Microsoft. All rights reserved.
//    
//    Apache 2.0 License
//    
//    You may obtain a copy of the License at
//    http://www.apache.org/licenses/LICENSE-2.0
//    
//    Unless required by applicable law or agreed to in writing, software 
//    distributed under the License is distributed on an "AS IS" BASIS, 
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or 
//    implied. See the License for the specific language governing 
//    permissions and limitations under the License.
//
//*********************************************************//

using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio;
using Microsoft.VisualStudio.Shell.Interop;

namespace TestUtilities.Mocks {
    public class MockSettingsStore : IVsSettingsStore {
        private readonly List<Tuple<string, string, object>> Settings = new List<Tuple<string, string, object>>();

        public void AddSetting(string path, string name, object value) {
            if (string.IsNullOrEmpty(path)) {
                path = Settings.Last().Item1;
            } else if (path.StartsWith(" ")) {
                path = Settings.Last().Item1 + "\\" + path.TrimStart();
            }

            Settings.Add(Tuple.Create(path, name, value));
        }

        public void Clear() {
            Settings.Clear();
        }

        private int FindValue<T>(string collectionPath, string propertyName, ref T value) {
            var tuple = Settings.FirstOrDefault(t => t.Item1 == collectionPath && t.Item2 == propertyName);
            if (tuple == null) {
                return VSConstants.S_FALSE;
            }
            if (!(tuple.Item3 is T)) {
                return VSConstants.E_INVALIDARG;
            }
            value = (T)tuple.Item3;
            return VSConstants.S_OK;
        }

        public int CollectionExists(string collectionPath, out int pfExists) {
            var collectionSeq = collectionPath.Split('\\');
            pfExists = Settings
                .Select(t => t.Item1.Split('\\'))
                .Where(seq => seq.Length >= collectionSeq.Length)
                .Any(seq => seq.Take(collectionSeq.Length).SequenceEqual(collectionSeq)) ? 1 : 0;
            return VSConstants.S_OK;
        }

        public int GetBinary(string collectionPath, string propertyName, uint byteLength, byte[] pBytes = null, uint[] actualByteLength = null) {
            throw new NotImplementedException();
        }

        public int GetBool(string collectionPath, string propertyName, out int value) {
            return GetBoolOrDefault(collectionPath, propertyName, 0, out value);
        }

        public int GetBoolOrDefault(string collectionPath, string propertyName, int defaultValue, out int value) {
            bool res = (defaultValue != 0);
            var hr = FindValue(collectionPath, propertyName, ref res);
            value = res ? 1 : 0;
            return hr;
        }

        public int GetInt(string collectionPath, string propertyName, out int value) {
            value = 0;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetInt64(string collectionPath, string propertyName, out long value) {
            value = 0;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetInt64OrDefault(string collectionPath, string propertyName, long defaultValue, out long value) {
            value = defaultValue;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetIntOrDefault(string collectionPath, string propertyName, int defaultValue, out int value) {
            value = defaultValue;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetLastWriteTime(string collectionPath, SYSTEMTIME[] lastWriteTime) {
            throw new NotImplementedException();
        }

        public int GetPropertyCount(string collectionPath, out uint propertyCount) {
            propertyCount = (uint)Settings.Count(t => t.Item1 == collectionPath);
            return VSConstants.S_OK;
        }

        public int GetPropertyName(string collectionPath, uint index, out string propertyName) {
            try {
                propertyName = Settings.Where(t => t.Item1 == collectionPath).ElementAt((int)index).Item2;
                return VSConstants.S_OK;
            } catch (InvalidOperationException) {
                propertyName = null;
                return VSConstants.E_INVALIDARG;
            }
        }

        public int GetPropertyType(string collectionPath, string propertyName, out uint type) {
            throw new NotImplementedException();
        }

        public int GetString(string collectionPath, string propertyName, out string value) {
            value = null;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetStringOrDefault(string collectionPath, string propertyName, string defaultValue, out string value) {
            value = defaultValue;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetSubCollectionCount(string collectionPath, out uint subCollectionCount) {
            var collectionSeq = collectionPath.Split('\\');
            if (!Settings
                .Select(t => t.Item1.Split('\\'))
                .Where(seq => seq.Length >= collectionSeq.Length)
                .Where(seq => seq.Take(collectionSeq.Length).SequenceEqual(collectionSeq))
                .Any()) {
                subCollectionCount = 0;
                return VSConstants.E_INVALIDARG;
            }
            
            subCollectionCount = (uint)Settings
                .Select(t => t.Item1.Split('\\'))
                .Where(seq => seq.Length > collectionSeq.Length)
                .Where(seq => seq.Take(collectionSeq.Length).SequenceEqual(collectionSeq))
                .Select(seq => seq[collectionSeq.Length])
                .Distinct()
                .Count();
            return VSConstants.S_OK;
        }

        public int GetSubCollectionName(string collectionPath, uint index, out string subCollectionName) {
            var collectionSeq = collectionPath.Split('\\');
            if (!Settings
                .Select(t => t.Item1.Split('\\'))
                .Where(seq => seq.Length >= collectionSeq.Length)
                .Where(seq => seq.Take(collectionSeq.Length).SequenceEqual(collectionSeq))
                .Any()) {
                subCollectionName = null;
                return VSConstants.E_INVALIDARG;
            }

            subCollectionName = Settings
                .Select(t => t.Item1.Split('\\'))
                .Where(seq => seq.Length > collectionSeq.Length)
                .Where(seq => seq.Take(collectionSeq.Length).SequenceEqual(collectionSeq))
                .Select(seq => seq[collectionSeq.Length])
                .Distinct()
                .ElementAt((int)index);
            return VSConstants.S_OK;
        }

        public int GetUnsignedInt(string collectionPath, string propertyName, out uint value) {
            value = 0;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetUnsignedInt64(string collectionPath, string propertyName, out ulong value) {
            value = 0;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetUnsignedInt64OrDefault(string collectionPath, string propertyName, ulong defaultValue, out ulong value) {
            value = defaultValue;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int GetUnsignedIntOrDefault(string collectionPath, string propertyName, uint defaultValue, out uint value) {
            value = defaultValue;
            return FindValue(collectionPath, propertyName, ref value);
        }

        public int PropertyExists(string collectionPath, string propertyName, out int pfExists) {
            pfExists = Settings.Any(t => t.Item1 == collectionPath && t.Item2 == propertyName) ? 1 : 0;
            return VSConstants.S_OK;
        }
    }
}
